"""
Fritter, the Frame-Rate IndependenT TimEr tRee.

This is a library for scheduling timed events in the very near or very far
future.
"""

__version__ = "0.0.8"
"The current version of the Fritter library."
